import  hashlib
import  io
from    atomicwrites import atomic_write
from    wurlitzer    import pipes, STDOUT

from    . import cells as c, utils, evaluate, image, mpl

from    .traceback import Traceback

class Hasher:
    def __init__(self):
        self.m = hashlib.sha256()

    def update(self, cell):
        if not cell.hashable: return
        code = cell.code()
        self.m.update(bytes(code, 'utf-8'))

    def digest(self):
        return self.m.digest()

class Notebook:
    def __init__(self, name = '', auto_capture = False, debug = False):
        self.name = name
        self.auto_capture = auto_capture
        self.debug = debug

        self.incoming = []
        self.current  = 0

        self.cells = []

        self.g = {'__name__' : '__main__'}
        self.l = self.g
        self.m = Hasher()

    def add(self, cells):
        self.incoming += cells

    def skip(self, location, output = lambda x: None):
        while self.current < location:
            cell = self.incoming[self.current]
            self.append(cell, output)
            self.current += 1
            if type(cell) is c.CodeCell:
                if cell.skippable:
                    self.m.update(cell)
                else:
                    self.execute(cell, output)

        cell = self.incoming[self.current]
        assert type(cell) is c.CheckpointCell
        self.g = cell.load()
        self.l = self.g
        self.append(cell, output)
        self.current += 1

    def move_all_incoming(self):
        while self.current < len(self.incoming):
            cell = self.incoming[self.current]
            self.cells.append(cell)
            self.current += 1

    def next_cell(self):
        if self.current < len(self.incoming):
            return self.incoming[self.current]
        else:
            return None

    def execute(self, cell, output, info = lambda *args: None):
        cell_id = len(self.cells) - 1

        # figure out the range of cells that belong to the current cell
        begin = self.current
        while type(self.next_cell()) in [c.Blanks, c.VariableCell, c.OutputCell]:
            self.current += 1
        # leave the last Blanks to MarkdownCell or BreakCell
        if type(self.next_cell()) in [c.MarkdownCell, c.BreakCell] and type(self.incoming[self.current-1]) is c.Blanks:
            self.current -= 1

        # reset back
        end = self.current
        self.current = begin

        # skip blanks, if any
        while type(self.next_cell()) is c.Blanks:
            if self.current >= end:
                break
            self.append(self.next_cell())
            self.current += 1

        self.m.update(cell)

        if type(self.next_cell()) is c.VariableCell:
            if self.next_cell().expected(self.m.digest()):
                info(f"Previous code cell not evaluated, [affirm]loading[/affirm] [variables]{self.next_cell().variables.strip()}[/variables] instead")
                self.next_cell().load(self.l)
                self.append(self.next_cell(), output)
                self.current += 1
                return

        with utils.captured_passthrough() as out:
            mpl.figures = out

            result = evaluate.exec_eval(cell.code(), self.g, self.l, name = f"{self.name}:{cell_id}")
            if result is not None:
                out.write(result.__repr__() + '\n')
                if self.auto_capture and image.is_new_mpl_available():
                    out.append_png(image.save_mpl_png())

        if type(self.next_cell()) is c.VariableCell:
            try:
                self.next_cell().dump(self.m.digest(), self.l)
            except:
                info(f"[error]Failed[/error] to save varialbes [variables]{self.next_cell().variables.strip()}[/variables], [affirm]skipping[/affirm]")
                tb = Traceback(self, self.debug)
                info(tb, block = True)
                info(f"[affirm]continuing[/affirm]")
            self.append(self.next_cell(), output)
            self.current += 1

        # skip the next output cell
        self.skip_next_output()

        self.append(c.OutputCell(out), lambda cell: None)
        print()

    def skip_next_output(self):
        if type(self.next_cell()) is c.VariableCell:
            self.append(self.next_cell())
            self.current += 1
        if type(self.next_cell()) is c.OutputCell:
            self.current += 1

    def process(self, output, info = lambda *args: None):
        while self.current < len(self.incoming):
            cell = self.incoming[self.current]
            self.current += 1

            self.append(cell, output)

            if type(cell) is c.CodeCell:
                self.execute(cell, output, info)
            elif type(cell) is c.VariableCell:
                info(f"[warn]Warning:[/warn] isolated variable cell [variables]{cell.variables.strip()}[/variables]; result will never be used")
                try:
                    cell.dump(self.m.digest(), self.l)
                except:
                    info(f"[error]Failed[/error] to save variables [variables]{cell.variables.strip()}[/variables], [affirm]skipping[/affirm]")
                    tb = Traceback(self, self.debug)
                    info(tb, block = True)
                    info(f"[affirm]continuing[/affirm]")
            elif type(cell) is c.CheckpointCell:
                try:
                    cell.dump(self.m.digest(), self.l)
                except:
                    info(f"[error]Failed[/error] to save state in checkpoint cell, [affirm]skipping[/affirm]")
                    tb = Traceback(self, self.debug)
                    info(tb, block = True)
                    info(f"[affirm]continuing[/affirm]")

    def append(self, cell, output = lambda x: None):
        self.cells.append(cell)
        if not self.debug and not cell.display(): return
        output(cell)

    def rehash(self):
        while self.current < len(self.incoming):
            cell = self.incoming[self.current]
            self.current += 1

            if type(cell) is c.CodeCell:
                self.m.update(cell)
            elif type(cell) is c.VariableCell or type(cell) is c.CheckpointCell:
                cell.replace_hash(self.m.digest())

            self.append(cell)

    def save(self, fn):
        with atomic_write(fn, mode='w', overwrite=True) as f:
            for i,cell in enumerate(self.cells):
                for line in cell.save():
                    f.write(line)
                if i != len(self.cells) - 1 and not line.endswith('\n'):
                    f.write('\n')

    def find_checkpoint(self):
        m = Hasher()
        last = None
        for i,cell in enumerate(self.incoming):
            if type(cell) is c.CodeCell:
                m.update(cell)
            elif type(cell) is c.CheckpointCell:
                if cell.expected(m.digest()):
                    last = i
                else:
                    break
        return last



from matplotlib._pylab_helpers import Gcf
from matplotlib.backend_bases import (_Backend, FigureCanvasBase, FigureManagerBase)
from matplotlib.backends.backend_agg import FigureCanvasAgg

import io

figures = None

class SaturnFigureManager(FigureManagerBase):
    def show(self):
        with io.BytesIO() as buf:
            self.canvas.figure.savefig(buf, format='png')
            figures.append_png(buf.getvalue())

@_Backend.export
class SaturnBackend(_Backend):
    FigureCanvas  = FigureCanvasAgg
    FigureManager = SaturnFigureManager

    def show(*args, **kwargs):
        _Backend.show(*args, **kwargs)
        Gcf.destroy_all()



import matplotlib
matplotlib.use("module://saturn_notebook.mpl")

import matplotlib.pyplot as plt
import io
import os
import sys
from base64 import standard_b64encode
from . import utils
from more_itertools import peekable

enabled = os.environ['TERM'] == 'xterm-kitty' and sys.stdout.isatty()      # TODO: find a better way
seen = []

def is_new_mpl_available():
    global seen
    axes = plt.gcf().get_axes()
    for ax in axes:
        if ax not in seen:
            seen = axes
            return True
    seen = axes
    return False

def save_mpl_png():
    buf = io.BytesIO()
    plt.savefig(buf, format='png')
    return buf.getvalue()

def show_png(buf, fd = None):
    if enabled:
        icat(buf, fd)

def icat(buf, fd = None):
    def serialize_gr_command(cmd, payload=None):
       cmd = ','.join('{}={}'.format(k, v) for k, v in cmd.items())
       ans = io.BytesIO()
       w = ans.write
       w(b'\033_G'), w(cmd.encode('ascii'))
       if payload:
          w(b';')
          w(payload)
       w(b'\033\\')
       return ans.getbuffer()

    if fd is None:
        write = sys.stdout.buffer.write
    else:
        write = lambda x: os.write(fd, x)

    pchunk = peekable(utils.chunkstring(standard_b64encode(buf), 4096))
    cmd = { 'a': 'T', 'f': 100 }
    for chunk in pchunk:
        cmd['m'] = 1 if pchunk else 0
        write(serialize_gr_command(cmd, chunk))
        cmd.clear()
    write(b'\n')



import sys, os, io, contextlib
from  wurlitzer         import pipes, STDOUT, Wurlitzer, _default_encoding
from  . import image

# From: https://stackoverflow.com/a/18854817/44738
def chunkstring(string, length):
    return (string[0+i:length+i] for i in range(0, len(string), length))

class CompositeIO:
    def __init__(self):
        self.outer = [io.StringIO()]
        self.outfd = None

    def empty(self):
        if len(self.outer) > 1: return False
        self.outer[0].seek(0, io.SEEK_END)
        if self.outer[0].tell() > 0: return False
        return True

    def set_outfd(self, outfd):
        self.outfd = outfd

    def __iter__(self):
        return iter(self.outer)

    def write(self, s):
        self.outer[-1].write(s)
        if self.outfd is not None:
            os.write(self.outfd, s.encode(_default_encoding))

    def append_png(self, buf):
        self.outer.append(buf)
        self.outer.append(io.StringIO())
        if self.outfd is not None:
            image.show_png(buf, self.outfd)

@contextlib.contextmanager
def captured_passthrough():
    out = CompositeIO()
    w = Wurlitzer(stdout=out, stderr=STDOUT, encoding=_default_encoding)
    with w:
        out.set_outfd(w._save_fds['stdout'])
        yield out

def collapse_carriage_return(lines):
    result = []
    for line in lines:
        while True:
            up = line.find('\x1b[A')
            cr = line.find('\r')

            if up != -1:
                result.pop()
                line = line[up+1:]
            elif cr != -1:
                line = line[cr+1:]
            else:
                result.append(line)
                break

    return result




import  os
import  argh
from    rich.console import Console
from    rich.rule    import Rule
from    rich.theme   import Theme
from    rich.style   import Style

import  io
from    atomicwrites import atomic_write
from    more_itertools import peekable

from    .           import cells as c, notebook
from    .repl       import PythonReplWithExecute
from    .image      import show_png

try:
    from mpi4py import MPI
    root = MPI.COMM_WORLD.Get_rank() == 0
    using_mpi = MPI.COMM_WORLD.Get_size() > 1
except ImportError:
    root = True
    using_mpi = False

# workaround for a bug in OpenMPI (or anything else that screws up the terminal size);
# see https://github.com/willmcgugan/rich/issues/127
import  shutil
width = None if shutil.get_terminal_size().columns != 0 else 80

theme = Theme({
    "variables":    Style.parse("yellow"),
    "warn":         Style.parse("yellow"),
    "affirm":       Style.parse("green"),
    "error":        Style.parse("red"),
    "cell-name":    Style.parse("yellow"),
})

console = Console(width = width, theme = theme)

def info(*args, block = False, **kw):
    if root:
        if not block:
            console.print(Rule(*args, **kw))
        else:
            console.print(*args, **kw)

def show_console(cell, rule = False, verbose = False, no_show = False):
    if rule:
        console.print(Rule(cell.type_name() if verbose else ''))

    if not no_show:
        cell.show_console(console)

    if not cell.empty() and not rule:
        console.print()

def show_html(cell, f):
    cell.show_html(f)

katex_preamble = r"""
<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/katex@0.16.3/dist/katex.min.css" integrity="sha384-Juol1FqnotbkyZUT5Z7gUPjQ9gzlwCENvUZTpQBAPxtusdwFLRy382PSDx5UUJ4/" crossorigin="anonymous">
<script defer src="https://cdn.jsdelivr.net/npm/katex@0.16.3/dist/katex.min.js" integrity="sha384-97gW6UIJxnlKemYavrqDHSX3SiygeOwIZhwyOKRfSaf0JWKRVj9hLASHgFTzT+0O" crossorigin="anonymous"></script>
<script defer src="https://cdn.jsdelivr.net/npm/katex@0.16.3/dist/contrib/auto-render.min.js" integrity="sha384-+VBxd3r6XgURycqtZ117nYw44OOcIax56Z4dCRWbxyPt0Koah1uHoK0o4+/RRE05" crossorigin="anonymous"></script>
<script>
    document.addEventListener("DOMContentLoaded", function() {
        renderMathInElement(document.body, {
          // customised options
          // • auto-render specific keys, e.g.:
          delimiters: [
              {left: '$$', right: '$$', display: true},
              {left: '$', right: '$', display: false},
              {left: '\\(', right: '\\)', display: false},
              {left: '\\[', right: '\\]', display: true}
          ],
          // • rendering keys, e.g.:
          throwOnError : false
        });
    });
</script>
<script src="https://cdn.jsdelivr.net/npm/katex@0.16.3/dist/contrib/copy-tex.min.js" integrity="sha384-ww/583aHhxWkz5DEVn6OKtNiIaLi2iBRNZXfJRiY1Ai7tnJ9UXpEsyvOITVpTl4A" crossorigin="anonymous"></script>
"""

def show(fn: "input notebook",
         html: "save HTML to a file" = '',
         katex: "include KaTeX in HTML output" = False,
         debug: "show debugging information" = False):
    """Show the contents of the notebook, without evaluating."""
    with open(fn) as f:
        cells = c.parse(f, show_only = True)
    _show(cells, html, katex, debug)

def _show(cells, html, katex, debug):
    output   = lambda cell: show_console(cell, rule = debug, verbose = debug)

    if html:
        f_html = open(html, 'w')
        output = lambda cell: show_html(cell, f_html)

        f_html.write('<!DOCTYPE html>\n')
        f_html.write('<html>\n')
        f_html.write('<head>\n')
        f_html.write('<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/water.css@2/out/light.css">\n')
        if katex:
            f_html.write(katex_preamble)
        f_html.write('<style>\n')
        f_html.write(c.HtmlFormatter().get_style_defs('.highlight'))
        f_html.write('</style>\n')
        f_html.write('</head>\n')
        f_html.write('<body>\n')


    for i,cell in enumerate(cells):
        if not cell.display(): continue
        output(cell)

    if html:
        f_html.write('</body>\n')
        f_html.write('</html>\n')

@argh.arg('outfn', nargs='?')
@argh.arg('-n', '--dry-run')
@argh.arg('-i', '--interactive')
def run(infn: "input notebook",
        outfn: "output notebook (if empty, input modified in place)",
        clean: "run from scratch, ignoring checkpoints" = False,
        auto_capture: "automatically capture images" = False,
        debug: "show debugging information" = False,
        dry_run: "don't save the processed notebook" = False,
        only_root_output: "suppress output everywhere but rank 0 (for MPI)" = False,
        interactive: "run REPL after the notebook is processed" = False):
    """Run the notebook."""
    if not outfn:
        outfn = infn

    if os.path.exists(infn):
        with open(infn) as f:
            cells = c.parse(f)
    else:
        cells = []

    def output(cell):
        if root or (not only_root_output and type(cell) is c.OutputCell):
            show_console(cell, rule = debug, verbose = debug)

    nb = notebook.Notebook(name = infn, auto_capture = auto_capture)
    nb.add(cells)

    if not clean:
        checkpoint = nb.find_checkpoint()
        if checkpoint is not None:
            info(f"Skipping to checkpoint {checkpoint}", style='magenta')
            nb.skip(checkpoint, output)
            info('Resuming', style="magenta")

    try:
        nb.process(output, info)

        if interactive:
            run_repl(nb, output, outfn, dry_run)
    except SystemExit:
        info("Caught SystemExit")
        nb.move_all_incoming()
    except:
        info("Caught exception, aborting")
        from .traceback import Traceback
        tb = Traceback(nb, debug = debug, width = 80)

        console_tb = Console(record = True, width = 80, theme = theme)
        console_tb.print(tb)

        nb.skip_next_output()
        nb.append(c.OutputCell.from_string(console_tb.export_text()))
        nb.move_all_incoming()

    if not dry_run and root:
        nb.save(outfn)


def run_repl(nb, output, outfn = '', dry_run = True):
    if using_mpi:
        comm = MPI.COMM_WORLD

    def execute_line(line):
        if using_mpi and root:
            line = comm.bcast(line, root = 0)
        blank = c.Blanks()
        blank.append('\n')
        cells = [blank] + c.parse(io.StringIO(line))
        nb.add(cells)
        nb.process(output)

    if not root:
        while True:
            line = comm.bcast('', root = 0)
            if not line: return
            execute_line(line)

    repl = PythonReplWithExecute(
        execute = execute_line,
        get_globals=lambda: nb.g,
        get_locals=lambda: nb.l,
        vi_mode=False,
        history_filename=None,
        startup_paths=None,
    )

    # Add the code cells to history
    for cell in nb.cells:
        for line in cell.repl_history():
            repl.history.append_string(line)

    @repl.add_key_binding('c-o')
    def _(event):
        if not dry_run and root:
            nb.save(outfn)

    repl.run()

    if using_mpi:
        comm.bcast('', root = 0)

@argh.arg('outfn', nargs='?')
def clean(infn: "input notebook",
          outfn: "output notebook (if empty, input modified in place)",
          strip_output: "also strip all output" = False):
    """Remove all binary data from the notebook."""
    if not outfn:
        outfn = infn

    if os.path.exists(infn):
        with atomic_write(outfn, mode='w', overwrite=True) as of:
            with open(infn) as f:
                pf = peekable(f)
                for line in pf:
                    if strip_output and line.startswith('#o>'): continue
                    if line.startswith('#o> png'): continue
                    if line.startswith('#chk>') and line.strip() != '#chk>':
                        of.write('#chk>\n')
                        while pf and pf.peek().startswith('#chk>'):
                            next(pf)
                        continue
                    if line.startswith('#var>'):
                        # Keep the first line, but skip all subsequent lines
                        while pf and pf.peek().startswith('#var>'):
                            next(pf)
                    of.write(line)

@argh.arg('i',   nargs='?', type=int)
@argh.arg('out', nargs='?')
def image(infn: "input notebook", i: "image index", out: "output PNG filename"):
    """Extract an image from the notebook."""
    if i is not None and not out:
        console.print("Must specify output filename, if image is specified")
        return

    with open(infn) as f:
        cells = c.parse(f, show_only = True)

    count = 0
    for cell in cells:
        if type(cell) is not c.OutputCell: continue

        for x in cell.composite_:
            if type(x) is io.StringIO: continue

            if i is None:
                print(f"{count}:")
                show_png(x)
            elif i == count:
                with open(out, 'wb') as f:
                    f.write(x)

            count += 1

def version():
    """Show version of Saturn and its dependencies."""
    from importlib_metadata import version as ver
    print(f"Saturn {ver('saturn_notebook')}")
    for dep in ['wurlitzer', 'rich', 'ptpython',
                'dill', 'markdown', 'atomicwrites',
                'pygments', 'more_itertools', 'matplotlib', 'nbformat']:
        print(f"   {dep} {ver(dep)}")

@argh.arg('outfn', nargs='?')
def convert(infn: "Jupyter notebook",
            outfn: "output notebook (if empty, show the cells instead)",
            version: "notebook version" = 4,
            html: "save HTML to a file" = '',
            katex: "include KaTeX in HTML output" = False,
            debug: "show debugging information" = False):
    """Convert a Jupyter notebook into a Saturn notebook."""
    import nbformat
    jnb = nbformat.read(infn, as_version=version)

    header = c.CodeCell()
    header.lines_ = ['# vim: ft=python foldmethod=marker foldlevel=0\n']
    cells = [header]

    for jcell in jnb.cells:
        if jcell['cell_type'] == 'markdown':
            cell = c.MarkdownCell()
            cell.lines_ = [' ' + line + '\n' if len(line) else '\n' for line in jcell['source'].split('\n')]
            if type(cells[-1]) is not c.Blanks:
                cells.append(c.Blanks.create(1))
            cells.append(cell)
            cells.append(c.Blanks.create(1))
        elif jcell['cell_type'] == 'code':
            if type(cells[-1]) is c.CodeCell:
                cells.append(c.Blanks.create(1))
                cells.append(c.BreakCell())
                cells.append(c.Blanks.create(1))
            cell = c.CodeCell()
            cell.lines_ = [line + '\n' for line in jcell['source'].split('\n')]
            cells.append(cell)

            for out in jcell['outputs']:
                if out['output_type'] == 'stream':
                    cell = c.OutputCell.from_string(out['text'])
                    cells.append(cell)
                elif out['output_type'] == 'display_data' and 'image/png' in out['data']:
                    cell = c.OutputCell()
                    png_content = out['data']['image/png']
                    cell.composite_.append_png(c.base64.b64decode(png_content))
                    cells.append(cell)
                else:
                    info('Unrecognized output type', style="magenta")
        else:
            info('Unrecognized cell type', style="magenta")

    if not outfn:
        _show(cells, html, katex, debug)
    else:
        nb = notebook.Notebook(name = outfn)
        nb.add(cells)
        nb.move_all_incoming()
        nb.save(outfn)


@argh.arg('outfn', nargs='?')
def rehash(infn: "input notebook",
           outfn: "output notebook (if empty, input modified in place)"):
    """Rehash all the code cells, updating the hashes stored with checkpoints and variable cells. (advanced)"""
    if not outfn:
        outfn = infn

    with open(infn) as f:
        cells = c.parse(f)

    nb = notebook.Notebook(name = infn)
    nb.add(cells)
    nb.rehash()

    nb.save(outfn)

def main():
    argh.dispatch_commands([show, run, clean, image, version, convert, rehash])

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['saturn_notebook']

package_data = \
{'': ['*']}

install_requires = \
['Markdown>=3.3.6,<4.0.0',
 'Pygments>=2.12.0,<3.0.0',
 'argh>=0.26.2,<0.27.0',
 'atomicwrites>=1.4.0,<2.0.0',
 'dill>=0.3.4,<0.4.0',
 'importlib-metadata>=4.11.3,<5.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'more-itertools>=8.12.0,<9.0.0',
 'nbformat>=5.7.0,<6.0.0',
 'ptpython>=3.0.20,<4.0.0',
 'rich>=12.3.0,<13.0.0',
 'wurlitzer>=3.0.2,<4.0.0']

entry_points = \
{'console_scripts': ['saturn = saturn_notebook.__main__:main']}

setup_kwargs = {
    'name': 'saturn-notebook',
    'version': '1.0.6',
    'description': 'Plain-text Python notebooks with checkpointing',
    'long_description': "# Saturn\n\n[Screenshots](#screenshots)\n\n## Features\n\n* Plain-text format. Notebooks are regular Python files. Different types of\n  cells are comments with special formatting.\n\n* Checkpoints. Special checkpoint cells allow to save the state of the session\n  or individual variables.\n\n* Terminal graphics support. When using\n  [kitty](https://sw.kovidgoyal.net/kitty/) terminal (or in principle anything\n  that supports its [graphics protocol](https://sw.kovidgoyal.net/kitty/graphics-protocol.html))\n  matplotlib figures are rendered inline in the terminal.\n\n* MPI awareness. When running under MPI, only rank 0 will write out the\n  modified notebook. The REPL will take input on rank 0 and broadcast to other\n  ranks. It's also possible to suppress output from all ranks other than 0.\n\n## Commands and options\n\n* `saturn show notebook.py`\n\n  Display the notebook in the terminal. No computation is performed. Optional\n  `--html OUTPUT.html` flag will produce HTML output. Use `-k, --katex` flag to\n  embed [KaTeX](https://katex.org/) header into the HTML.\n\n  `saturn show notebook.py --html notebook.html -k`\n\n* `saturn run notebook.py [output.py]`\n\n  Execute a Python notebook, either modifying it in place, or saving the result\n  into a new notebook `output.py`.\n\n  * `-c, --clean`: run from scratch, ignoring the checkpoints.\n  * `-a, --auto-capture`: automatically capture matplotlib figures, without `show()`.\n  * `-i`, `--interactive`:\n    drop into REPL (using [ptpython](https://github.com/prompt-toolkit/ptpython))\n    after all the cells are processed; the results of the REPL interaction will\n    be added to the notebook.\n  * `-n, --dry-run`: don't save the result.\n  * `--only-root-output`: under MPI, suppress output from all ranks other than 0.\n\n* `saturn clean notebook.py [output.py]`\n\n  Remove all binary data from the notebook. Useful for getting rid of large\n  checkpoints.\n\n* `saturn image notebook.py [i out.png]`\n\n  Save `i`-th image from `notebook.py` into `out.png`. If the last two\n  arguments are omitted, show all the images in the notebook together with\n  their indices.\n\n* `saturn version`\n\n  Show version of saturn and its dependencies.\n\n* `saturn convert notebook.ipynb [notebook.py]`\n\n  Convert a Jupyter notebook into a Saturn notebook. If the output name\n  `notebook.py` is missing, only show the Jupyter notebook. Optional\n  `--html OUTPUT.html` flag will produce HTML output.\n\n\n## Cell types\n\n* Markdown cells, prefix `#m>`\n\n  ```\n  #m> # Sample notebook\n  #m>\n  #m> Description using markdown **formatting**.\n  ```\n\n* Output cells `#o>`\n\n  There is not usually a reason to modify these by hand, they are filled by\n  Saturn with the output of code cells. If they contain PNG information, it's\n  base64-encoded and wrapped in `{{{` and `}}}` to allow automatic folding.\n\n  ```\n  #o> <matplotlib.image.AxesImage object at 0x114217550>\n  #o> png{{{\n  #o> pngiVBORw0KGgoAAAANSUhEUgAAA8AAAAHgCAYAAABq5QSEAAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAAP\n  ...\n  #o> pngGAAAgBQEMAAAACkIYAAAAFL4v5JTyvRQ4v1eAAAAAElFTkSuQmCC\n  #o> png}}}\n  ```\n\n  In Vim with `foldmethod=marker`:\n  ```\n  #o> <matplotlib.image.AxesImage object at 0x114217550>\n  +--135 lines: o> png--------------------------------------------------\n  ```\n\n* Checkpoint cells `#chk>`\n\n  These indicate locations, where the code should checkpoint. Checkpointing\n  serializes the session, which is stored base64-encoded in the same cell. The\n  cell also stores the hash of the previous code blocks, and the checkpoint is\n  valid if the prior code blocks haven't changed. By default saturn will resume\n  from the last valid checkpoint. Same folding markers (`{{{` and `}}}`) are used.\n\n  ```\n  #chk>{{{\n  #chk>gANDIJZCQePiVH9SX7wVtBfgrDpcgWu5HUFFiFEeyNF9sVjFcQB9cQEoWAwAAABfX2J1aWx0aW5zX19x\n  ...\n  #chk>wAyP55wdmz+qIkdBjBrYP3EjdHEkYnWGcSUu\n  #chk>}}}\n  ```\n\n  In Vim with `foldmethod=marker`:\n  ```\n  +-- 36 lines: chk>----------------------------------------------------\n  ```\n\n* Variable cells `#var> x,y,z`\n\n  These cells save only the value of the specified variables (which is useful\n  if the full checkpoint is too big). If all the previous code cells haven't\n  changed, the cell's saved content is loaded into the specified variables and\n  the previous code cell is not evaluated.\n\n* Break cells `#---#`\n\n  These are used to break code cells that don't have any other type of a cell\n  between them.\n\n* Code cells\n\n  All contiguous lines, not marked as one of the above, are grouped together\n  into code cells.\n\n* Non-skippable code cells `#no-skip#`\n\n  Adding this line anywhere within the code cell will indicate that it\n  shouldn't be skipped, even if we are restarting from a checkpoint. This is\n  useful, for example, if a cell is modifying `sys.path`, which won't be\n  captured in a checkpoint.\n\n* Non-hashable code cells `#no-hash#`\n\n  Adding this line anywhere within the code cell will indicate that it\n  shouldn't be hashed, meaning that changing this cell (or removing it\n  entirely) won't invalidate the checkpoints below. This should be used only\n  with cells that don't change any variables, e.g., purely output or plotting\n  cells.\n\n## Vim support\n\nAll the binary (non-human-readable) cell content is wrapped in `{{{`, `}}}`\nmarkers. Adding the following comment to the notebook, ensures that Vim starts\nwith all the binary content folded away.\n\n```\n# vim: foldmethod=marker foldlevel=0\n```\n\n## Screenshots\n\nRunning [samples/simple.py](https://github.com/mrzv/saturn/blob/master/samples/simple.py):\n\n* First run performs full computation and saves the checkpoint, as well as the figure output.\n\n![First run](https://github.com/mrzv/saturn/raw/master/resources/screenshots/simple-first-run.png)\n\n* Second run resumes from the checkpoint, since no code before it has changed.\n\n![Second run](https://github.com/mrzv/saturn/raw/master/resources/screenshots/simple-second-run.png)\n\n* Vim folds the binary content.\n\n![Vim folding](https://github.com/mrzv/saturn/raw/master/resources/screenshots/simple-vim.png)\n",
    'author': 'Dmitriy Morozov',
    'author_email': 'dmitriy@mrzv.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mrzv/saturn',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import ast
import sys

# From: https://stackoverflow.com/questions/33908794/get-value-of-last-expression-in-exec-call
# exec, but return the value of the last expression
def exec_eval(script, globals=None, locals=None, name=''):
    '''Execute a script and return the value of the last expression'''
    stmts = list(ast.iter_child_nodes(ast.parse(script)))
    if not stmts:
        return None
    if isinstance(stmts[-1], ast.Expr):
        # the last one is an expression and we will try to return the results
        # so we first execute the previous statements
        if len(stmts) > 1:
            if sys.version_info >= (3, 8):
                mod = ast.Module(stmts[:-1], [])
            else:
                mod = ast.Module(stmts[:-1])
            exec(compile(mod, filename=name, mode="exec"), globals, locals)
        # then we eval the last one
        return eval(compile(ast.Expression(body=stmts[-1].value), filename=name, mode="eval"), globals, locals)
    else:
        # otherwise we just execute the entire code
        return exec(compile(script, filename=name, mode="exec"), globals, locals)

def eval_expression(expr, locals_):
    return eval(expr, locals_, locals_)



from rich.markdown import Markdown
from rich.syntax   import Syntax
from rich.text     import Text

from  itertools import chain
import base64
import io
import dill
from  more_itertools    import peekable

import html
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import HtmlFormatter

import markdown

from . import utils
from . import image
from . import evaluate

class Cell:
    def __init__(self):
        self.lines_ = []

    def append(self, line):
        prefix = self.__class__._prefix
        line = line[len(prefix):]       # eat the prefix
        self.lines_.append(line)

    def save(self):
        prefix = self.__class__._prefix
        return [prefix + line for line in self.lines_]

    def parse(self):            # called after all the lines have been read
        pass

    def lines(self):
        return ''.join(self.lines_)

    def repl_history(self):
        return self.save()

    @classmethod
    def display(cls):
        return True

    def empty(self):
        return not self.__class__.display()

    @classmethod
    def identify(cls, line):
        return line.startswith(cls._prefix)

    def type_name(self):
        if hasattr(self.__class__, '_name'):
            return self.__class__._name
        else:
            return self.__class__.__name__

    def __repr__(self):
        return '== ' + self.__class__.__name__ + '\n' + self.lines()

    def __rich__(self):
        return '[cell-name]== ' + self.__class__.__name__ + ' ==[/cell-name]\n' + self.lines()

    def show_console(self, console):
        console.print(self)

    def show_html(self, f):
        f.write(self._render_html())
        f.write('\n')

    def _render_html(self):
        return f"<div><h2>{self.__class__.__name__}</h2> <pre>{html.escape(self.lines())}</pre></div>"

class CodeCell(Cell):
    _prefix = ''
    _name   = 'Code'

    def __init__(self):
        super().__init__()
        self.skippable = True
        self.hashable  = True

    def __rich__(self):
        if not self.lines_:
            return
        return Syntax('>>> ' + '... '.join(self.lines_).strip(), 'python')

    def append(self, line):
        super().append(line)
        if line.startswith('#no-skip#'):
            self.skippable = False
        if line.startswith('#no-hash#'):
            self.hashable = False

    def code(self):
        return self.lines()

    def _render_html(self):
        return f"<div class='code'>{highlight(self.code(), PythonLexer(), HtmlFormatter())}</div>"

class MarkdownCell(Cell):
    _prefix = '#m>'
    _name   = 'Markdown'

    def __rich__(self):
        return Markdown(self.lines())

    def _render_html(self):
        return "<div class='markdown'>" + markdown.markdown(''.join(line[1:] if line[0] == ' ' else line for line in self.lines_)) + "</div>"

class OutputCell(Cell):
    _prefix = '#o> '
    _name   = 'Output'

    def __init__(self, composite_ = None):
        super().__init__()
        if composite_ is None:
            self.composite_ = utils.CompositeIO()
        else:
            self.composite_ = composite_

    def repl_history(self):
        return []

    @staticmethod
    def from_string(s):
        cell = OutputCell()
        cell.composite_.write(s)
        return cell

    def parse(self):
        super().parse()
        pl = peekable(self.lines_)
        for line in pl:
            if not line.startswith('png'):
                self.composite_.write(line)
            else:
                png_content = [line[3:]]
                while pl and pl.peek().startswith('png') and not pl.peek().startswith('png{{{'):
                    png_content.append(next(pl)[3:])
                png_content = ''.join(png_content)
                self.composite_.append_png(base64.b64decode(png_content))
        self.lines_.clear()

    def save(self):
        lines_ = []
        for x in self.composite_:
            if type(x) is io.StringIO:
                x.seek(0)
                lines_ += [self._prefix + line for line in utils.collapse_carriage_return(x)]
            else:
                content = base64.b64encode(x).decode('ascii')
                lines_ += [self._prefix + 'png' + line + '\n' for line in chunk(content, 80, markers = True)]
        return lines_

    def empty(self):
        return self.composite_.empty()

    def show_console(self, console):
        for x in self.composite_:
            if type(x) is io.StringIO:
                if x.getvalue():
                    console.print(Text(x.getvalue()))
            else:
                image.show_png(x)

    def _render_html(self):
        result = ""
        for x in self.composite_:
            if type(x) is io.StringIO:
                result += f"<div class='output'><pre>{html.escape(x.getvalue())}</pre>\n"
            else:
                result += f'<img src="data:image/png;base64,{base64.b64encode(x).decode("ascii")}"/>\n'
        result += "</div>"
        return result

class BreakCell(Cell):
    _prefix = '#---#'

    @classmethod
    def display(cls):
        return False

    def save(self):
        return [self._prefix + '\n']

class CheckpointCell(Cell):
    _prefix = '#chk>'

    def expected(self, h):
        if self.expected_hash() == None:
            return False
        else:
            return self.expected_hash() == h

    def expected_hash(self):
        if not hasattr(self, '_expected'):
            self._expected = dill.load(self._content)
        return self._expected

    @classmethod
    def display(cls):
        return False

    def repl_history(self):
        return []

    def parse(self):
        if all(l.strip() == '' for l in self.lines_):
            self.lines_ = []

        if not self.lines_:
            self._expected = None
            return

        content = self.lines()
        content = base64.b64decode(content)
        self._content = io.BytesIO(content)

    def load(self):
        h = self.expected_hash()    # just to make sure it's been parsed
        assert h != None
        if not hasattr(self, '_locals'):
            self._locals = dill.load(self._content)
        return self._locals

    def dump(self, running, locals_):
        try:
            content = io.BytesIO()
            dill.dump(running, content)
            dill.dump(locals_, content)
            content = base64.b64encode(content.getvalue()).decode('ascii')
            self.lines_ = [line + '\n' for line in chunk(content, 80, markers = True)]
        except:
            self.lines_ = ['']      # to keep the blank checkpoint cell
            raise

    def replace_hash(self, running):
        if self.expected_hash() == None: return
        content = io.BytesIO()
        dill.dump(running, content)
        content.write(self._content.read())
        content = base64.b64encode(content.getvalue()).decode('ascii')
        self.lines_ = [line + '\n' for line in chunk(content, 80, markers = True)]

class VariableCell(CheckpointCell):
    _prefix = '#var>'

    def parse(self):
        self.variables = self.lines_[0]
        self.lines_ = self.lines_[1:]

        super().parse()

    def load(self, locals_):
        locals_['_var_cell_load'] = dill.load(self._content)
        evaluate.exec_eval(self.variables.strip() + ' = _var_cell_load', locals_, locals_)
        del locals_['_var_cell_load']

    def dump(self, running, locals_):
        content = io.BytesIO()
        dill.dump(running, content)
        dill.dump(evaluate.eval_expression(self.variables, locals_), content)
        content = base64.b64encode(content.getvalue()).decode('ascii')
        self.lines_ = [line + '\n' for line in chunk(content, 80, markers = True)]

    def header(self):
        return self.__class__._prefix + self.variables

    def save(self):
        lines_ = super().save()
        return [self.header()] + lines_

    def repl_history(self):
        return [self.header()]

# Captures blank lines between cells
class Blanks(Cell):
    _prefix = ''

    def show_console(self, console):
        pass

    def show_html(self, f):
        pass

    @classmethod
    def display(cls):
        return False

    @staticmethod
    def create(n):
        cell = Blanks()
        cell.lines_ = ['\n']*n
        return cell

cell_types = [MarkdownCell, OutputCell, BreakCell, CheckpointCell, VariableCell]

def identify(line):
    for Type in chain(cell_types, [CodeCell]):        # CodeCell matches everything, so comes last
        if Type.identify(line):
            return Type

def chunk(content, width, markers = False):
    chunking = utils.chunkstring(content, 80)
    if markers:
        chunking = chain(['{{{'], chunking, ['}}}'])
    return chunking

def parse(f, show_only = False):
    cells = []
    def cells_append(cell):
        if len(cells) > 0:
            cells[-1].parse()
        cells.append(cell)

    p = peekable(f)
    for line in p:
        # agglomerate empty lines into Blanks and either store as such or return to the CodeCell, if in the middle of one
        if not line.strip():
            blank = Blanks()
            blank.append(line)
            while p and not p.peek().strip():
                blank.append(next(p))

            if p and len(cells) > 0 and type(cells[-1]) is CodeCell and identify(p.peek()) is CodeCell:
                for line in blank.lines_:
                    cells[-1].append(line)
            else:
                cells_append(blank)

            continue

        Type = identify(line)

        if show_only and not Type.display(): continue     # skip check-point cells in show mode

        if len(cells) == 0 or type(cells[-1]) is not Type:
            cells_append(Type())
        cells[-1].append(line)

    if len(cells) > 0:
        cells[-1].parse()

    return cells
